"""High-level index access functions."""

import gflags

from commoncrawlindex import pbtree
from commoncrawlindex import s3

FLAGS = gflags.FLAGS

gflags.DEFINE_string(
  'index_path',
  ('s3://aws-publicdatasets/common-crawl/projects/url-index/'
   'url-index.1356128792'),
  'Path to the URL index to use.  Can be an s3:// URI or a local file path.',
  short_name='i')


def open_index_stream(path=None):
  """Opens a index stream for the index at a given path.

  Args:
    path(string): A local file path or an a3:// URI to an index file.
      Default is the value of --index_path.
  """
  path = path or FLAGS.index_path
  if s3.is_s3_uri(path):
    s3_conn = s3.get_s3_connection()
    stream = s3.BotoMap(s3_conn, path)
  else:
    stream = open(path, 'rb')
  return stream


def open_index_reader(path=None):
  """Opens and returns a reader for the index at a given path.

  Args:
    path(string): A local file path or an s3:// URI to an index
      file. Default is the value of --index_path.
  """
  path = path or FLAGS.index_path
  stream = open_index_stream(path=path)
  return pbtree.open_pbtree_reader(stream)



# Copyright [2012] triv.io, Scott Robertson
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import bisect
import struct
from tempfile import TemporaryFile, SpooledTemporaryFile
from cStringIO import StringIO
import itertools

from commoncrawlindex import prefix

MB = 1024 ** 2
DISK_BLOCK_SIZE = 1024 * 4

OFFSET_FMT = '<I'
OFFSET_SIZE = struct.calcsize(OFFSET_FMT)


class Error(Exception):
  pass


class PBTreeWriter(object):
  """Constructs a disk based prefixed btree for a scalar value."""
  def __init__(self, stream, block_size=MB, terminator='\0',
               value_format='<Q'):
    self.stream = stream
    if len(terminator) != 1:
      raise Error('terminator must be of length 1')
    self.value_format = value_format
    self.value_size = struct.calcsize(self.value_format)
    self.block_size = block_size
    self.last_key = ''
    self.data_segment = DataWriter(
      TemporaryFile(), block_size, terminator, self)
    self.index_segment = IndexWriter(stream, block_size, terminator)

  def pack_value(self, value):
    return struct.pack(self.value_format, value)

  def add(self, key, value):
    self.data_segment.add(key, value)
    self.last_key = key

  def on_new_block(self, key):
    prefix_key = prefix.significant(self.last_key, key)
    self.index_segment.add(0, prefix_key)

  def on_item_exceeds_block_size(self, key, unused_value):
    raise ValueError("key '%s' exceeds block size" % key)

  def commit(self):
    out = self.stream
    self.index_segment.finish()
    self.data_segment.finish()
    while True:
      bytes = self.data_segment.read(DISK_BLOCK_SIZE)
      if bytes == '':
        break
      else:
        out.write(bytes)

  def close(self):
    self.commit()
    self.stream.close()


class PBTreeSequenceWriter(PBTreeWriter):
  """Writes a sequence to a pbtree, it assumes that an appropriate
  value_format is specified and that every sequence passed to add() is
  the same length.
  """
  def pack_value(self, value):
    # this assumes an appropriate value_format was specified
    return struct.pack(self.value_format, *value)


class PBTreeDictWriter(PBTreeWriter):
  """Writes a dictionary to a pbtree."""
  def __init__(self, stream, item_keys, **options):
    self.item_keys = item_keys
    super(PBTreeDictWriter, self).__init__(stream, **options)

  def pack_value(self, dict):
    value = [dict[k] for k in self.item_keys]
    return struct.pack(self.value_format, *value)


class PBTreeReader(object):
  """Reads a prefixed btree."""
  @classmethod
  def parse(cls, stream, block_size):
    while True:
      block = stream.read(block_size)
      if block == '':
        break
      for offset, key in IndexBlockReader(block):
        yield offset, key

  def __init__(self, mmap, terminator='\0', value_format='<Q'):
    self.mmap = mmap
    #self.mmap.seek(0)
    self.terminator = terminator
    self.value_format = value_format
    self.value_size = struct.calcsize(self.value_format)
    self.header_fmt = '<II'
    self.header_size = struct.calcsize(self.header_fmt)
    self.block_size, self.index_block_size = self.fetch_header()

  def fetch(self, start, end):
    return self.mmap[start:end]

  def fetch_header(self):
    return struct.unpack(self.header_fmt, self.fetch(0, self.header_size))

  def block(self, block_number):
    """Returns the block for given block number."""
    offset = self.block_offset(block_number)
    block = self.fetch(offset, offset + self.block_size)
    return IndexBlockReader(block)

  def block_offset(self, block_number):
    return self.header_size + (self.block_size * block_number)

  def count_levels(self):
    """Return the number of 'levels' in the index. This number
    represents how many seeks have to be preformed before finding the
    starting data block.
    """
    block_number = 0
    block = self.block(block_number)
    levels = 1
    while True:
      buffer = StringIO(block.data)
      next_block_number = block.read_offset(buffer)
      if next_block_number < self.index_block_size:
        block_number = next_block_number
        block = self.block(block_number)
        levels += 1
      else:
        # it's the start of the data segments.
        return levels

  def expected_location(self, key):
    """Given a key, return the expected starting location for the key
    in the data segment. The return value is only the location where
    the key "should" be not neccesarily where it is. This enables
    range queries.
    """
    if key == '':
      start_block = self.index_block_size
      return self.block_offset(start_block)
    starting_block = self.find_starting_data_block(key)
    offset = self.block_offset(starting_block)
    data = self.fetch(offset, offset + self.block_size)
    # linear scan through the block, looking for the position of the
    # stored key that is greater than the given key.
    start = 0
    while True:
      pos = data.find(self.terminator, start)
      if pos == -1:
        # can this really happen?
        return len(data)
      stored = data[start:pos]
      if key <= stored:
        return start + offset
      else:
        start = pos + 1 + self.value_size

  def find_starting_data_block(self, key):
    """Return the number of the first data block where the key should
    be found.
    """
    block_number = 0
    block = self.block(block_number)
    while True:
      next_block_number = block.find(key)
      if next_block_number < self.index_block_size:
        block_number = next_block_number
        block = self.block(block_number)
      else:
        # it's the start of the data segments,
        return next_block_number

  def keys(self, prefix=''):
    return list(self.keyiter(prefix))

  def keyiter(self, prefix=''):
    for key, unused_value in self.itemsiter(prefix):
      if key.startswith(prefix):
        yield key

  def values(self, prefix=''):
    return list(self.valueiter(prefix))

  def valueiter(self, prefix=''):
    for key, value in self.itemsiter(prefix):
      if key.startswith(prefix):
        yield value

  def items(self, prefix=''):
    return list(self.itemsiter(prefix))

  def itemsiter(self, prefix):
    starting_block = self.find_starting_data_block(prefix)
    blocks = self.blockiter(starting_block)
    # skip over keys in the first block
    first_block = blocks.next()
    for key, value in itertools.dropwhile(
      lambda (k, v): not k.startswith(prefix), self.dataiter(first_block)):
      if not key.startswith(prefix):
        return
      yield key, value
    for block in blocks:
      for key, value in self.dataiter(block):
        if not key.startswith(prefix):
          return
        else:
          yield key, value

  def parse_value(self, bytes):
    """Returns a scalar."""
    return struct.unpack(
      self.value_format,
      bytes
    )[0]

  def blockiter(self, block_number):
    """Iterate over blocks starting with the given block_number."""
    while True:
      offset = self.block_offset(block_number)
      block = self.mmap[offset:offset + self.block_size]
      if block:
        yield block
        block_number += 1
      else:
        break

  def dataiter(self, block):
    for key, value in DataBlockReader(block, self.value_size, self.terminator):
      yield key, self.parse_value(value)


class PBTreeDictReader(PBTreeReader):
  def __init__(self, stream, item_keys, **options):
    self.item_keys = item_keys
    super(PBTreeDictReader, self).__init__(stream, **options)

  def parse_value(self, bytes):
    """Returns a dictionary."""
    return dict(zip(
      self.item_keys,
      struct.unpack(self.value_format, bytes)
    ))


####
# Lower level constructs whose functionality has been seperated out to make
# usabale within map/reduce constructs
###

class DataWriter(object):
  def __init__(self, stream, block_size, terminator, delegate):
    self.finalized = False
    self.delegate = delegate
    self.block_size = block_size
    self.remaining = self.block_size
    self.terminator = terminator
    self.term_length = len(self.terminator)
    self.stream = stream
    self.write_buffer = bytearray()

  def add(self, key, value):
    packet = self.delegate.pack_value(value)
    size = len(key) + self.term_length + len(packet)
    if size > self.block_size:
      self.delegate.on_item_exceeds_block_size(key, value)
      return
    if size > self.remaining:
      self.write_buffer.extend(self.terminator * self.remaining)
      self.stream.write(self.write_buffer)
      del self.write_buffer[:]
      self.remaining = self.block_size
      self.delegate.on_new_block(key)
    self.write_buffer.extend(key)
    self.write_buffer.extend(self.terminator)
    self.write_buffer.extend(packet)
    self.remaining -= size

  def close(self):
    if not self.finalized:
      self.finish()
    self.stream.close()

  def finish(self):
    if self.write_buffer:
      self.write_buffer.extend(self.terminator * self.remaining)
      self.stream.write(self.write_buffer)
    del self.write_buffer
    self.delegate = None
    self.stream.seek(0)
    self.finalized = True

  def read(self, bytes=None):
    """Returns bytes written to the stream.

    It is an error to call this method prior to calling DataWriter.finish().
    """
    return self.stream.read(bytes)


class IndexWriter(object):
  def __init__(self, stream, block_size, terminator, pointer_format='<I'):
    self.stream = stream
    self.block_size = block_size
    self.terminator = terminator
    self.term_size = len(terminator)
    self.pointer_format = pointer_format
    self.pointer_size = struct.calcsize(pointer_format)
    self.indexes = []
    self.push_index()

  def add(self, level, key):
    """Adds the key to the given index level."""
    size = len(key) + self.term_size + self.pointer_size
    stream, pointers, remaining = self.indexes[level]

    if size > remaining:
      # pad the rest with null bytes
      stream.write(self.terminator * remaining)
      assert stream.tell() % self.block_size == 0
      # start the block off with the offset
      stream.write(struct.pack(self.pointer_format, pointers))
      next_level = level + 1
      if next_level > len(self.indexes) - 1:
        self.push_index()
      self.add(next_level, key)
      remaining = self.block_size - self.pointer_size

    pointers += 1
    stream.write(key)
    stream.write(self.terminator)
    stream.write(struct.pack(self.pointer_format, pointers))
    remaining = remaining - size
    self.indexes[level] = stream, pointers, remaining

  def push_index(self):
    stream = SpooledTemporaryFile(max_size=20 * MB)
    pointers = 0
    stream.write(struct.pack(OFFSET_FMT, pointers))
    self.indexes.append([
      stream, pointers, self.block_size - self.pointer_size
    ])

  def finish(self):
    out = self.stream
    blocks_written = 0
    out.write(struct.pack(OFFSET_FMT, self.block_size))
    # blocks in the index
    out.write(struct.pack(OFFSET_FMT, 0))

    for stream, unused_pointers, remaining in reversed(self.indexes):
      # pad the stream
      stream.write(self.terminator * remaining)
      level_length = stream.tell()
      assert level_length % self.block_size == 0
      blocks_to_write = (level_length / self.block_size)
      stream.seek(0)
      # loop through each pointer and key writing
      for o, key in PBTreeReader.parse(stream, self.block_size):
        out.write(
          struct.pack(
            self.pointer_format, o + blocks_written + blocks_to_write))
        # note: the last key in the block returned by the reader will
        # be all null bytes pads
        out.write(key)
      blocks_written += blocks_to_write
      stream.close()

    out.seek(OFFSET_SIZE)
    out.write(struct.pack(OFFSET_FMT, blocks_written))
    out.seek(0, 2)  # move to the end of the file
    self.blocks_written = blocks_written

  def close(self):
    self.finish()


class DataBlockReader(object):
  def __init__(self, bytes, value_size, terminator='\0'):
    self.bytes = bytes
    self.terminator = terminator
    self.value_size = value_size

  def __iter__(self):
    block = self.bytes
    start = 0
    while True:
      pos = block.find(self.terminator, start)
      if pos == -1:
        break
      key = block[start:pos]
      start = pos + 1
      if key == '':
        return
      else:
        value = block[start:start + self.value_size]
        start += self.value_size
        yield key, value


class IndexBlockReader(object):
  def __init__(self, data):
    self.data = data

  def __iter__(self):
    end_of_block = False
    buffer = StringIO(self.data)
    while not end_of_block:
      offset = self.read_offset(buffer)
      key = self.read_key(buffer)
      # near the end of the block, eveything from here on should be null bytes
      if key == '':
        # ended on block boundry
        end_of_block = True
      elif key == '\0':
        # nothing but pad bytes left
        end_of_block = True
        key += self.read_rest_of_block(buffer)
      yield offset, key

  def find(self, key):
    pointers = []
    prefixes = []
    for pointer, prefix in self:
      pointers.append(pointer)
      prefixes.append(prefix)
    # discard padding
    prefixes.pop()
    index = bisect.bisect(prefixes, key)
    return pointers[index]

  def read_offset(self, buffer):
    bytes = buffer.read(OFFSET_SIZE)
    # bytes should never be empty when reading an offset if so the file
    # corrupt
    return struct.unpack(OFFSET_FMT, bytes)[0]

  def read_key(self, buffer):
    buff = bytearray()
    while True:
      c = buffer.read(1)
      if c == '':
        if len(buff):
          raise RuntimeError('EOF found when string was expected')
        else:
          break
      elif c == '\0':
        buff.append(c)
        break
      else:
        buff.append(c)
    return str(buff)

  def read_rest_of_block(self, buffer):
    return buffer.read()


def open_pbtree_reader(stream):
  reader = PBTreeDictReader(
    stream,
    value_format='<QQIQI',
    item_keys=(
      'arcSourceSegmentId',
      'arcFileDate',
      'arcFilePartition',
      'arcFileOffset',
      'compressedSize'))
  return reader



# <data segment> :: = <item> (<item> | ( {<null>} <null>))
# <item> ::= <key> <null> <location pointer>

#
#

import unittest

from commoncrawlindex import prefix


class TestIndexMapCase(unittest.TestCase):
  def test(self):
    class P():
      pass
    params = P()
    results = []
    for partition_number, input in enumerate(inputs):
      params.last_key = None
      params.partition_number = partition_number
      for block in input:
        for item in map_block(iter(block), params):
          results.append(item)
    self.assertSequenceEqual(results, final)


file1 = [
  [
    'key01',
    'key02',
    'key03a',
    'key03ac',
  ],
  [
    'key03bc'
    'key06',
    'key07',
    'key08z',
  ],
  [
    'key08zafz'
    'key10',
    'key11',
    'key12',
  ],
]

file2 = [
  [
    'key13feee',
    'key14',
    'key16',
    'key16a',
  ],
  [
    'key16b'
    'key18',
    'key19',
    'key20',
  ]
]


final = (
  (0, 'key01'),
  (0, 'key03b'),
  (0, 'key08za'),
  (1, 'key13feee'),
  (1, 'key16b')
)

inputs = [file1, file2]


def map_block(block, params):
  # yield first item and last

  first_key = block.next()
  assert first_key.find('\0') == -1

  if params.last_key is None:
    yield params.partition_number, first_key
  else:
    yield (params.partition_number,
           prefix.significant(params.last_key, first_key))

  second_to_last = None
  for key in block:
    if not key.startswith('\0'):
      second_to_last = key

  if second_to_last is not None:
    params.last_key = second_to_last
  else:
    params.last_key = first_key



"""Fetches a URL from the Common Crawl dataset.

Usage:
  %s [options] <reversed URL prefix>...
"""

import gzip
import StringIO
import sys

import gflags

from commoncrawlindex import s3
from commoncrawlindex import index

FLAGS = gflags.FLAGS

gflags.DEFINE_boolean(
  'output_to_file',
  False,
  'Write each fetched URL to a file named like the URL.',
  short_name='O')
gflags.DEFINE_boolean(
  'compress',
  False,
  'Keep the URL contents gzipped.',
  short_name='C')


_S3_URI_TMPL = (
  's3://aws-publicdatasets/common-crawl/parse-output/segment/'
  '{arcSourceSegmentId}/{arcFileDate}_{arcFilePartition}.arc.gz')


def arc_file(s3_conn, info, decompress=True):
  """Reads an ARC file (see
  http://commoncrawl.org/data/accessing-the-data/).
  """
  s3_path = _S3_URI_TMPL.format(**info)
  bucket_name, key_name = s3.parse_s3_uri(s3_path)
  bucket = s3_conn.lookup(bucket_name)
  key = bucket.lookup(key_name)
  start = info['arcFileOffset']
  end = start + info['compressedSize'] - 1
  headers = {'Range': 'bytes={}-{}'.format(start, end)}
  contents = key.get_contents_as_string(headers=headers)
  if decompress:
    chunk = StringIO.StringIO(contents)
    return gzip.GzipFile(fileobj=chunk).read()
  else:
    return contents


def url_to_filename(url):
  """Converts a URL to a valid filename."""
  return url.replace('/', '_')


def main():
  try:
    argv = FLAGS(sys.argv)
  except gflags.FlagsError, e:
    sys.stderr.write('Error: %s\n%s%s\n' % (
        e,
        sys.modules[__name__].__doc__.replace('%s', sys.argv[0]),
        FLAGS))
    sys.exit(2)
  if len(argv) < 2:
    sys.stderr.write('Error: Wrong number of arguments.\n')
    sys.exit(1)
  index_reader = index.open_index_reader()
  s3_conn = s3.get_s3_connection()
  try:
    for url_prefix in argv[1:]:
      for url, d in index_reader.itemsiter(url_prefix):
        sys.stderr.write('Fetching %s\n' % (url,))
        contents = arc_file(s3_conn, d, decompress=(not FLAGS.compress))
        if FLAGS.output_to_file:
          filename = url_to_filename(url)
          if FLAGS.compress:
            filename = filename + '.gz'
          with open(filename, 'wb') as f:
            f.write(contents)
        else:
          print contents
  except KeyboardInterrupt:
    pass


if __name__ == '__main__':
  main()



# Copyright 2012 Triv.io, Scott Robertson
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import itertools


def commonlen(s1, s2):
  """Returns the length of the common prefix."""
  # given "hi", "hip"
  # izip_longest("hi", "hip") -> ('h','h'), ('i','i'), (None, 'p')
  # enumerate -> (0,('h','h')), (1,('i','i')), (2,(None, 'p'))
  # dropwhile(lambda (i,(x,y)): x == 5 -> (2,(None,'p')) ...
  try:
    return itertools.dropwhile(
      lambda (i, (x, y)): x == y, enumerate(zip(s1, s2))).next()[0]
  except StopIteration:
    # strings are identical return the len of one of them
    return len(s1)


def common(s1, s2):
  """Returns the common prefix."""
  cl = commonlen(s1, s2)
  return s2[:cl]


def significant(s1, s2):
  """Given two strings s1 and s2 and assuming s2 > s1, returns the
  character that make s2 greater.
  """
  cl = commonlen(s1, s2)
  return s2[:cl + 1]



# Copyright [2012] [Triv.io, Scott Robertson]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import urlparse


def reversehost(url):
  """Reverses the 'netloc' of a URL and mobes the protocol to the end.
  For example, http://www.example.com/foo -> com.example.www/foo:http
  """
  url = urlparse.urlsplit(str(url))
  netloc = url.netloc.split(':')
  host = netloc[0]
  if len(netloc) == 2:
    port = ':' + netloc[1]
  else:
    port = ''
  # reverse the host
  host = '.'.join(reversed(host.split('.')))
  return (
    host +
    url.path +
    (('?' + url.query) if url.query else '') +
    port +
    ':' + url.scheme
  )



# Copyright [2012] [Triv.io, Scott Robertson]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import re
import struct


from triv.io import datasources
from .pbtree import PBTreeDictWriter, DataWriter, IndexWriter, DataBlockReader

BLOCK_SIZE   = 2**16
VALUE_FORMAT = "<QQIQ"
VALUE_SIZE   = struct.calcsize(VALUE_FORMAT)

ITEM_KEYS=(
  'arcSourceSegmentId',
  'arcFileDate',
  'arcFileParition',
  'arcFileOffset'
)


@datasources.write_mimetype('application/vnd.commoncrawl.org.url_index')
def output_stream(stream):
  return PBTreeDictWriter(stream,
    value_format=VALUE_FORMAT,
    block_size =BLOCK_SIZE,
    item_keys=ITEM_KEYS
  )


@datasources.write_mimetype('application/vnd.commoncrawl.org.data_segment')
def output_stream(stream):
  class DictPacker(object):
    def __init__(self, item_keys, value_format):
      self.item_keys = item_keys
      self.value_format = value_format
      self.block_count = 0

    def pack_value(self, dict):
      value = [dict[k] for k in self.item_keys]
      return struct.pack(self.value_format, *value)

    def on_new_block(self, key):
      self.block_count += 1

    def on_item_exceeds_block_size(self, key,value):
      print ""

  return DataWriter(stream, BLOCK_SIZE, '\0', DictPacker(ITEM_KEYS, VALUE_FORMAT))

@datasources.write_mimetype('application/vnd.commoncrawl.org.index_segment')
def output_stream(stream):
  return IndexWriter(stream, block_size=BLOCK_SIZE, terminator='\0')



@datasources.read_mimetype('application/vnd.commoncrawl.org.data_segment')
def data_block_reader(fd, size, url, params):
  part_number = "%04d" % int(re.search(':reduce:(\d+)\-',url).group(1))

  # read each block in the part, yielding the first key and part number

  params.last_key = None

  while True:
    block_buffer = fd.read(BLOCK_SIZE)
    if block_buffer == '':
      break
    block = iter(DataBlockReader(block_buffer, VALUE_SIZE))
    yield part_number, block



"""Prints a list of URLS from the Common Crawl Index that match a
prefix.

Usage:
  %s [options] <reversed URL prefix>
"""

import sys

import gflags

from commoncrawlindex import index

FLAGS = gflags.FLAGS

gflags.DEFINE_boolean(
  'print_metadata',
  False,
  'Print metadata for each URL.',
  short_name='m')


def main():
  try:
    argv = FLAGS(sys.argv)
  except gflags.FlagsError, e:
    sys.stderr.write('Error: %s\n%s%s\n' % (
        e,
        sys.modules[__name__].__doc__.replace('%s', sys.argv[0]),
        FLAGS))
    sys.exit(2)
  if len(argv) != 2:
    sys.stderr.write('Error: Wrong number of arguments.\n')
    sys.exit(1)
  index_reader = index.open_index_reader()
  try:
    for url, d in index_reader.itemsiter(argv[1]):
      if FLAGS.print_metadata:
        print url, d
      else:
        print url
  except KeyboardInterrupt:
    pass


if __name__ == '__main__':
  main()



"""High-level S3 functions."""

import re
import threading

import boto


class Error(Exception):
  pass


_S3_URI_RE = re.compile(r's3://([^/]+)/?(.*)')


def parse_s3_uri(s3_path):
  """Parses an s3:// URI into (bucket, key)."""
  m = _S3_URI_RE.match(s3_path)
  if not m:
    raise Error('%s is not a valid S3 URI' % (s3_path,))
  bucket = m.group(1)
  key_prefix = m.group(2)
  return (bucket, key_prefix)


def is_s3_uri(path):
  """Returns true if PATH is an S3 URI."""
  m = _S3_URI_RE.match(path)
  return not(not m)


g_s3_connections = threading.local()


def get_s3_connection():
  """Gets an S3 (boto) connection.

  Connections are cached per-thread.
  """
  global g_s3_connections
  if not hasattr(g_s3_connections, 'connection'):
    g_s3_connections.connection = boto.connect_s3(anon=True)
  return g_s3_connections.connection


class BotoMap(object):
  """A random-access interface to an S3 file."""
  def __init__(self, s3_conn, s3_uri):
    bucket_name, key_name = parse_s3_uri(s3_uri)
    bucket = s3_conn.lookup(bucket_name)
    self.key = bucket.lookup(key_name)
    self.block_size = 2 ** 16
    self.cached_block = -1

  def __getitem__(self, i):
    if isinstance(i, slice):
      start = i.start
      end = i.stop - 1
    else:
      start = i
      end = start + 1
    return self.fetch(start, end)

  def fetch(self, start, end):
    try:
      return self.key.get_contents_as_string(
        headers={'Range': 'bytes={0}-{1}'.format(start, end)}
      )
    except boto.exception.S3ResponseError, e:
      # invalid range, we've reached the end of the file.
      if e.status == 416:
        return ''



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
  from setuptools import setup
except ImportError:
  from distutils.core import setup


# Work around setuptools bug
# http://article.gmane.org/gmane.comp.python.peak/2509
import multiprocessing


PACKAGE_NAME = 'commoncrawlindex'
VERSION = '0.0.6'


settings = dict(
  name=PACKAGE_NAME,
  version=VERSION,
  description='Access to an index of Common Crawl URLs.',
  long_description='Access to an index of Common Crawl URLs',
  author='John Wiseman',
  author_email='jjwiseman@gmail.com',
  url='https://github.com/wiseman/common_crawl_index',
  packages=['commoncrawlindex', 'commoncrawlindex.cli'],
  test_suite='nose.collector',
  install_requires=[
    'boto==2.9.0',
    'python-gflags'
    ],
  tests_require=[
    'nose'
    ],
  license='MIT',
  classifiers=(
    # 'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Natural Language :: English',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    ),
  entry_points={
    'console_scripts': [
      'cci_lookup = commoncrawlindex.cli.cci_lookup:main',
      'cci_fetch = commoncrawlindex.cli.cci_fetch:main'
      ],
    },
  )


setup(**settings)



# Copyright [2012] [Triv.io, Scott Robertson]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import functools
import mmap
import os.path
import tempfile
import unittest

from nose.tools import eq_

from commoncrawlindex import pbtree

SORTED_URL_PATH = os.path.join(os.path.dirname(__file__), 'test_sorted_urls')


class TestIndex(unittest.TestCase):
  def test_btree_index(self):
    def data():
      """Returns an iterator of (url, linepos)."""
      with open(SORTED_URL_PATH, 'rb') as f:
        return ((url.strip(), pos) for pos, url in enumerate(f.readlines()))

    self.validate(
      pbtree.PBTreeWriter,
      pbtree.PBTreeReader,
      data(),
      prefix='http://natebeaty.com/',
      known_keys=[
        'http://natebeaty.com/illustration/4452349850',
        'http://natebeaty.com/illustration/4573016166',
        'http://natebeaty.com/illustration/4747271212',
        'http://natebeaty.com/illustration/4752986875',
      ],
      known_values=[
        1891,
        1892,
        1893,
        1894
      ]
    )

  def test_btree_dict_index(self):
    writer = functools.partial(
      pbtree.PBTreeDictWriter, item_keys=("key1", "key2"), value_format="<QI")
    reader = functools.partial(
      pbtree.PBTreeDictReader, item_keys=("key1", "key2"), value_format="<QI")

    def data():
      with open(SORTED_URL_PATH, 'rb') as f:
        for pos, url in enumerate(f):
          yield url.strip(), {'key1': pos, 'key2': pos}

    self.validate(
      writer,
      reader,
      data(),
      prefix='http://natebeaty.com/',
      known_keys=[
        'http://natebeaty.com/illustration/4452349850',
        'http://natebeaty.com/illustration/4573016166',
        'http://natebeaty.com/illustration/4747271212',
        'http://natebeaty.com/illustration/4752986875',
      ],
      known_values=[
        {'key1': 1891, 'key2': 1891},
        {'key1': 1892, 'key2': 1892},
        {'key1': 1893, 'key2': 1893},
        {'key1': 1894, 'key2': 1894}
      ]
    )

  def validate(self, writer, reader, data, prefix, known_keys, known_values):
    """Verify given writer produces content that can be read by the given
    reader and returns known keys and values for the given prefix.
    """
    temp = tempfile.NamedTemporaryFile(delete=False)
    index = writer(temp)

    keys_written = []
    for key, value in data:
      index.add(key, value)
      keys_written.append(key)
    index.close()

    ii = open(temp.name, 'r+')
    map = mmap.mmap(ii.fileno(), 0)
    index = reader(map)
    keys = index.keys(prefix)
    self.assertListEqual(keys, known_keys)

    values = index.values(prefix)
    self.assertListEqual(values, known_values)

    items = index.items(prefix)
    self.assertListEqual(items, zip(known_keys, known_values))
    self.assertListEqual(index.keys(), keys_written)



# Copyright 2012 Triv.io, Scott Robertson]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import tempfile
import unittest

from nose.tools import eq_

from commoncrawlindex import pbtree


class TestPBTree(unittest.TestCase):
  def test_btree_index(self):
    t = tempfile.TemporaryFile()
    tree = pbtree.PBTreeWriter(t)
    tree.add('blah', 1)
    #pbtree.commit()

    #t.seek(0)
    packet = tree.data_segment.write_buffer  # t.read()
    eq_(packet, 'blah\x00\x01\x00\x00\x00\x00\x00\x00\x00')

  def test_one_key_per_block_writer(self):
    # 2 pointers and a 1 byte string null terminated string = 10 bytes
    stream = tempfile.TemporaryFile()

    i = pbtree.IndexWriter(stream, block_size=10, terminator='\0')
    i.add(0, 'b')
    eq_(len(i.indexes), 1)

    i.add(0, 'c')
    eq_(len(i.indexes), 2)
    i.finish()

    stream.seek(0)
    packet = stream.read()
    eq_(len(packet), 30)

    root_block = packet[:10]
    eq_(root_block, '\x01\x00\x00\x00c\x00\x02\x00\x00\x00')

    block_1 = packet[10:20]
    eq_(block_1, '\x03\x00\x00\x00b\x00\x04\x00\x00\x00')

    block_2 = packet[20:]
    eq_(block_2, '\x04\x00\x00\x00c\x00\x05\x00\x00\x00')
